"""Custom middleware used by the pages application."""

import re

from django.conf import settings
from django.template.response import SimpleTemplateResponse

from cms.models import publication_manager, PublicationManagementError


class PublicationMiddleware(object):

    """Middleware that enables preview mode for admin users."""

    def __init__(self):
        """Initializes the PublicationMiddleware."""
        self.exclude_urls = [
            re.compile(url)
            for url in
            getattr(settings, "PUBLICATION_MIDDLEWARE_EXCLUDE_URLS", ())
        ]

    def process_request(self, request):
        """Starts preview mode, if available."""
        if not any(pattern.match(request.path_info[1:]) for pattern in self.exclude_urls):
            # See if preview mode is requested.
            try:
                preview_mode = bool(int(request.GET.get("preview", 0)))
            except ValueError:
                preview_mode = False
            # Only allow preview mode if the user is a logged in administrator.
            preview_mode = preview_mode and request.user.is_authenticated() and request.user.is_staff and request.user.is_active
            publication_manager.begin(not preview_mode)

    def process_response(self, request, response):
        """Cleans up after preview mode."""
        # Render the response if we're in a block of publication management.
        if publication_manager.select_published_active():
            if isinstance(response, SimpleTemplateResponse):
                response = response.render()
        # Clean up all blocks.
        while True:
            try:
                publication_manager.end()
            except PublicationManagementError:
                break
        # Carry on as normal.
        return response



"""Tests for the pages app."""

from django.test import TestCase
from django.contrib.contenttypes.models import ContentType

from cms import externals
from cms.apps.pages.models import Page, ContentBase


class TestPageContent(ContentBase):
    
    class Meta:
        app_label = "pages"


class PageEfficiencyTest(TestCase):
    
    def setUp(self):
        with externals.watson.context_manager("update_index")():
            content_type = ContentType.objects.get_for_model(TestPageContent)
            self.homepage = Page.objects.create(
                title = "Homepage",
                content_type = content_type,
            )
            TestPageContent.objects.create(
                page = self.homepage,
            )
            self.section = Page.objects.create(
                parent = self.homepage,
                title = "Section",
                content_type = content_type,
            )
            TestPageContent.objects.create(
                page = self.section,
            )
            self.subsection = Page.objects.create(
                parent = self.section,
                title = "Subsection",
                content_type = content_type,
            )
            TestPageContent.objects.create(
                page = self.subsection,
            )
            self.subsubsection = Page.objects.create(
                parent = self.subsection,
                title = "Subsubsection",
                content_type = content_type,
            )
            TestPageContent.objects.create(
                page = self.subsubsection,
            )
        
    def testChildPrefetching(self):
        # Make sure that prefetching works to two levels deep.
        with self.assertNumQueries(3):
            homepage = Page.objects.get_homepage()
        with self.assertNumQueries(0):
            subsection = homepage.children[0].children[0]
        self.assertEqual(subsection.title, "Subsection")
        with self.assertNumQueries(0):
            subsection = homepage.navigation[0].navigation[0]
        self.assertEqual(subsection.title, "Subsection")
        # Make sure that, beyond this, it doesn't go  pathalogical.
        with self.assertNumQueries(1):
            subsubsection = subsection.children[0]
        self.assertEqual(subsubsection.title, "Subsubsection")
        with self.assertNumQueries(0):
            subsubsection = subsection.children[0]
        self.assertEqual(subsubsection.title, "Subsubsection")


"""Admin settings for the static media management application."""

import os
from functools import partial

from django.contrib import admin
from django.contrib.admin.views.main import IS_POPUP_VAR
from django.contrib.staticfiles.storage import staticfiles_storage
from django.shortcuts import render
from django.template.defaultfilters import filesizeformat
from django.utils.text import Truncator

from sorl.thumbnail import get_thumbnail
from cms import permalinks, externals
from cms.apps.media.models import Label, File, Video


class LabelAdmin(admin.ModelAdmin):

    """Admin settings for Label models."""

    list_display = ("name",)

    search_fields = ("name",)


admin.site.register(Label, LabelAdmin)


class VideoAdmin(admin.ModelAdmin):

    def to_field_allowed(self, request, to_field):
        """
        This is a workaround for issue #552 which will raise a security
        exception in the media select popup with django 1.6.6.
        According to the release notes, this should be fixed by the
        yet (2014-09-22) unreleased 1.6.8, 1.5.11, 1.7.1.

        Details: https://code.djangoproject.com/ticket/23329#comment:11
        """

        if to_field == 'id':
            return True

        return super(VideoAdmin, self).to_field_allowed(request, to_field)

admin.site.register(Video, VideoAdmin)


# Different types of file.
AUDIO_FILE_ICON = "media/img/audio-x-generic.png"
DOCUMENT_FILE_ICON = "media/img/x-office-document.png"
SPREADSHEET_FILE_ICON = "media/img/x-office-spreadsheet.png"
TEXT_FILE_ICON = "media/img/text-x-generic.png"
IMAGE_FILE_ICON = "media/img/image-x-generic.png"
MOVIE_FILE_ICON = "media/img/video-x-generic.png"
UNKNOWN_FILE_ICON = "media/img/text-x-generic-template.png"

# Different types of recognised file extensions.
FILE_ICONS = {
    "mp3": AUDIO_FILE_ICON,
    "m4a": AUDIO_FILE_ICON,
    "wav": AUDIO_FILE_ICON,
    "doc": DOCUMENT_FILE_ICON,
    "odt": DOCUMENT_FILE_ICON,
    "pdf": DOCUMENT_FILE_ICON,
    "xls": SPREADSHEET_FILE_ICON,
    "txt": TEXT_FILE_ICON,
    "png": IMAGE_FILE_ICON,
    "gif": IMAGE_FILE_ICON,
    "jpg": IMAGE_FILE_ICON,
    "jpeg": IMAGE_FILE_ICON,
    "swf": MOVIE_FILE_ICON,
    "flv": MOVIE_FILE_ICON,
    "mp4": MOVIE_FILE_ICON,
    "mov": MOVIE_FILE_ICON,
    "wmv": MOVIE_FILE_ICON,
    'webm': MOVIE_FILE_ICON,
    'm4v': MOVIE_FILE_ICON,
}


class FileAdminBase(admin.ModelAdmin):

    """Admin settings for File models."""

    fieldsets = (
        (None, {
            "fields": ("title", "file",),
        },),
        ("Media management", {
            "fields": ("labels",),
            "classes": ("collapse",),
        },),
    )

    list_filter = ("labels",)

    search_fields = ("title",)

    list_display = ("get_preview", "get_title", "get_size",)

    change_list_template = "admin/media/file/change_list.html"

    filter_horizontal = ("labels",)

    def to_field_allowed(self, request, to_field):
        """
        This is a workaround for issue #552 which will raise a security
        exception in the media select popup with django 1.6.6.
        According to the release notes, this should be fixed by the
        yet (2014-09-22) unreleased 1.6.8, 1.5.11, 1.7.1.

        Details: https://code.djangoproject.com/ticket/23329#comment:11
        """

        if to_field == 'id':
            return True

        return super(FileAdminBase, self).to_field_allowed(request, to_field)

    # Customizations.

    def lookup_allowed(self, lookup, *args, **kwargs):
        """Allows the file iregex lookup needed by TinyMCE integration."""
        if lookup == "file__iregex":
            return True
        return super(FileAdminBase, self).lookup_allowed(lookup, *args, **kwargs)

    # Custom actions.

    def add_label_action(self, request, queryset, label):
        """Adds the label on the given queryset."""
        for file in queryset:
            file.labels.add(label)

    def remove_label_action(self, request, queryset, label):
        """Removes the label on the given queryset."""
        for file in queryset:
            file.labels.remove(label)

    def get_actions(self, request):
        """Generates the actions for assigning categories."""
        if IS_POPUP_VAR in request.GET:
            return []
        opts = self.model._meta
        verbose_name_plural = opts.verbose_name_plural
        actions = super(FileAdminBase, self).get_actions(request)
        # Add the dynamic labels.
        for label in Label.objects.all():
            # Add action.
            action_function = partial(self.__class__.add_label_action, label=label)
            action_description = u'Remove label %s from selected %s"' % (label.name, verbose_name_plural)
            action_name = action_description.lower().replace(" ", "_")
            actions[action_name] = (action_function, action_name, action_description)
            # Remove action.
            action_function = partial(self.__class__.remove_label_action, label=label)
            action_description = u'Remove label %s from selected %s"' % (label.name, verbose_name_plural)
            action_name = action_description.lower().replace(" ", "_")
            actions[action_name] = (action_function, action_name, action_description)
        return actions

    def remove_label(self, request, queryset):
        """Removes the label from selected files."""
        queryset.update(label=None)

    # Custom display routines.

    def get_label(self, obj):
        """Returns a pretty version of the label."""
        if obj.label:
            return obj.label.name
        return ""
    get_label.short_description = "label"
    get_label.admin_order_field = "label"

    def get_size(self, obj):
        """Returns the size of the media in a human-readable format."""
        try:
            return filesizeformat(obj.file.size)
        except OSError:
            return "0 bytes"
    get_size.short_description = "size"

    def get_preview(self, obj):
        """Generates a thumbnail of the image."""
        _, extension = os.path.splitext(obj.file.name)
        extension = extension.lower()[1:]
        icon = FILE_ICONS.get(extension, UNKNOWN_FILE_ICON)
        permalink = permalinks.create(obj)
        if icon == IMAGE_FILE_ICON:
            try:
                thumbnail = get_thumbnail(obj.file, '100x66', crop='center', quality=99)
            except IOError:
                pass
            else:
                return '<img cms:permalink="%s" src="%s" width="%s" height="%s" alt="" title="%s"/>' % (permalink, thumbnail.url, thumbnail.width, thumbnail.height, obj.title)
        else:
            icon = staticfiles_storage.url(icon)

        return '<img cms:permalink="%s" src="%s" width="66" height="66" alt="" title="%s"/>' % (permalink, icon, obj.title)
    get_preview.short_description = "preview"
    get_preview.allow_tags = True

    def get_title(self, obj):
        """Returns a truncated title of the object."""
        return Truncator(obj.title).words(8)
    get_title.short_description = "title"

    # Custom view logic.

    def response_add(self, request, obj, *args, **kwargs):
        """Returns the response for a successful add action."""
        if "_tinymce" in request.GET:
            context = {"permalink": permalinks.create(obj),
                       "title": obj.title}
            return render(request, "admin/media/file/filebrowser_add_success.html", context)
        return super(FileAdminBase, self).response_add(request, obj, *args, **kwargs)

    def changelist_view(self, request, extra_context=None):
        """Renders the change list."""
        context = {
            "changelist_template_parent": externals.reversion and "reversion/change_list.html" or "admin/change_list.html",
        }
        if extra_context:
            context.update(extra_context)
        return super(FileAdminBase, self).changelist_view(request, context)


# Renaming needed to allow inheritance to take place in this class without infinite recursion.
FileAdmin = FileAdminBase


if externals.reversion:
    class FileAdmin(FileAdmin, externals.reversion["admin.VersionMetaAdmin"]):
        list_display = FileAdmin.list_display + ("get_date_modified",)


if externals.watson:
    class FileAdmin(FileAdmin, externals.watson["admin.SearchAdmin"]):
        pass


admin.site.register(File, FileAdmin)



"""Permalink generation template tags."""


from django import template
from django.utils.html import escape

from cms import permalinks


register = template.Library()


@register.filter
def permalink(model):
    """Returns a permalink for the given model."""
    return permalinks.create(model)


@register.simple_tag(takes_context=True)
def permalink_absolute(context, model):
    request = context["request"]
    return escape(request.build_absolute_uri(permalinks.create(model)))


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('news', '0004_auto_20141014_1430'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='article',
            options={'ordering': ('-date',), 'permissions': (('can_approve_articles', 'Can approve articles'),)},
        ),
    ]



"""
Generic permalink generation utilities.

A permalink is a model and object id encoded into a URL.  It allows the model to
change it's absolute URL without breaking links.
"""


from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.views import shortcut
from django.core import urlresolvers
from django.core.exceptions import ImproperlyConfigured


__all__ = ("PermalinkError", "create", "resolve", "expand",)


class PermalinkError(Exception):

    """Exception thrown when an error occurs with a permalink."""


def create(obj):
    """Generates a permalink for the given object."""
    content_type = ContentType.objects.get_for_model(obj)
    content_type_id = content_type.id
    object_id = obj.pk
    kwargs = {"content_type_id": content_type_id,
              "object_id": object_id}
    return urlresolvers.reverse("permalink_redirect", kwargs=kwargs)


def resolve(permalink):
    """
    Resolves the given permalink into an object.

    Raises a PermalinkError if the URL is not a valid permalink. Raises an
    ObjectDoesNotExist if the referenced object does not exist.
    """
    # Attempt to resolve the URL.
    try:
        callback, _, callback_kwargs = urlresolvers.resolve(permalink)  # @UnusedVariable
    except (urlresolvers.Resolver404, TypeError):
        raise PermalinkError("'{}' is not a valid permalink.".format(permalink))
    # Check if the URL refers to a permalink.
    if callback != shortcut:
        raise PermalinkError("'{}' is not a valid permalink.".format(permalink))
    # Get the permalink attributes.
    try:
        content_type_id = callback_kwargs["content_type_id"]
        object_id = callback_kwargs["object_id"]
    except KeyError:
        raise ImproperlyConfigured("The permalink_redirect view should be configured using keyword arguments.")
    # Resolve the object.
    content_type = ContentType.objects.get_for_id(content_type_id)
    obj = content_type.get_object_for_this_type(id=object_id)
    return obj


def expand(permalink):
    """
    Expands the given permalink into a full URL.

    Raises a permalink error if the URL is not a valid permalink. Raises an
    ObjectDoesNotExist if the referenced object does not exist.
    """
    obj = resolve(permalink)
    return obj.get_absolute_url()





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(help_text=b'The title will be used as the default rollover text when this media is embedded in a web page.', max_length=200)),
                ('file', models.FileField(max_length=250, upload_to=b'uploads/files')),
            ],
            options={
                'ordering': ('title',),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Label',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=200)),
            ],
            options={
                'ordering': ('name',),
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='file',
            name='labels',
            field=models.ManyToManyField(help_text=b'Labels are used to help organise your media. They are not visible to users on your website.', to='media.Label', blank=True),
            preserve_default=True,
        ),
    ]



"""Context processors used by the pages application."""


def pages(request):
    """Adds the current page backend to the template."""
    context = {
        "pages": request.pages,
    }
    return context


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
import cms.apps.media.models


class Migration(migrations.Migration):

    dependencies = [
        ('media', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Video',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=200)),
                ('high_resolution_mp4', cms.apps.media.models.VideoFileRefField(related_name='+', on_delete=django.db.models.deletion.PROTECT, verbose_name=b'high resolution MP4', blank=True, to='media.File', null=True)),
                ('image', cms.apps.media.models.ImageRefField(related_name='+', on_delete=django.db.models.deletion.PROTECT, blank=True, to='media.File', null=True)),
                ('low_resolution_mp4', cms.apps.media.models.VideoFileRefField(related_name='+', on_delete=django.db.models.deletion.PROTECT, verbose_name=b'low resolution MP4', blank=True, to='media.File', null=True)),
                ('webm', cms.apps.media.models.VideoFileRefField(related_name='+', on_delete=django.db.models.deletion.PROTECT, verbose_name=b'WebM', blank=True, to='media.File', null=True)),
            ],
            options={
                'ordering': ('title',),
            },
            bases=(models.Model,),
        ),
    ]



"""Form widgets used by the CMS."""
import re
import string

from django import forms
from django.conf import settings
from django.contrib.auth.forms import PasswordChangeForm, AdminPasswordChangeForm
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe

from cms import debug

try:
    import json
except:
    from django.utils import simplejson as json


class HtmlWidget(forms.Textarea):

    """A textarea that is converted into a TinyMCE editor."""

    def __init__(self, *args, **kwargs):
        """Initializes the HtmlWidget."""
        self.richtext_settings = getattr(settings, "RICHTEXT_SETTINGS", {}).get(kwargs.pop("richtext_settings", "default"), {})
        super(HtmlWidget, self).__init__(*args, **kwargs)

    @debug.print_exc
    def get_media(self):
        """Returns the media used by the widget."""
        assets = [
            staticfiles_storage.url("cms/js/tiny_mce/tiny_mce.js"),
            staticfiles_storage.url("cms/js/jquery.cms.js"),
            staticfiles_storage.url("pages/js/jquery.cms.pages.js"),
            staticfiles_storage.url("media/js/jquery.cms.media.js"),
        ]
        return forms.Media(js=assets)

    media = property(
        get_media,
        doc = "The media used by the widget.",
    )

    def render(self, name, value, attrs=None):
        """Renders the widget."""

        # Add on the JS initializer.
        attrs = attrs or {}
        attrs['class'] = "tiny-mce"

        # Get the standard widget.
        html = super(HtmlWidget, self).render(name, value, attrs)

        try:
            element_id = attrs["id"]
        except KeyError:
            pass
        else:
            # Customize the config.
            richtext_settings = self.richtext_settings.copy()
            # Cache the asset URL.
            if "content_css" in richtext_settings:
                richtext_settings["content_css"] = staticfiles_storage.url(richtext_settings["content_css"])
            # Add in the initializer.
            settings_js = json.dumps(richtext_settings)
            html += u'<script>django.jQuery("#{element_id}").cms("htmlWidget",{settings_js})</script>'.format(
                element_id = element_id,
                settings_js = settings_js,
            )
        # All done!
        return mark_safe(html)


# Checks a string against some rules
def password_validation(password):

    errors = []

    if len(password) < 8:
        errors.append('Your password needs to be at least 8 characters long.')

    if password.lower() == password:
        errors.append('Your password needs include at least 1 uppercase character.')

    if password.upper() == password:
        errors.append('Your password needs include at least 1 lowercase character.')

    if not re.findall(r"[\d]", password):
        errors.append('Your password needs include at least 1 number.')

    if not re.findall(r"[{}]".format(re.escape(string.punctuation)), password):
        errors.append('Your password needs include at least 1 special character.')

    return errors


class CMSPasswordChangeForm(PasswordChangeForm):
    def clean_new_password1(self):
        password = self.cleaned_data.get('new_password1')
        if password_validation(password):
            raise ValidationError([
                ValidationError(error) for error in password_validation(password)
            ])
        return password


class CMSAdminPasswordChangeForm(AdminPasswordChangeForm):
    def clean_password1(self):
        password = self.cleaned_data.get('password1')
        if password_validation(password):
            raise ValidationError([
                ValidationError(error) for error in password_validation(password)
            ])
        return password



"""
Dynamic class loading routines.

Why isn't this in the Python standard library?  Why does Django re-implement it
a million times?  Most odd.
"""


def load_module(name):
    """Loads a module by name."""
    return __import__(name, {}, {}, [""])


def load_object(name):
    """Loads an object from a module."""
    module_name, object_name = name.rsplit(".", 1)
    module = load_module(module_name)
    return getattr(module, object_name)


"""Core models used by the CMS."""

from django.contrib.contenttypes.models import ContentType
from django.core import urlresolvers
from django.db import models, connection, transaction
from django.db.models import Q, F
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils import timezone

from cms import sitemaps, externals
from cms.models import PageBase, OnlineBaseManager, PageBaseSearchAdapter
from cms.models.managers import publication_manager


def get_default_page_parent():
    """Returns the default page parent."""
    try:
        return Page.objects.all()[0]
    except IndexError:
        return None


class PageManager(OnlineBaseManager):

    """Manager for Page objects."""

    def select_published(self, queryset, page_alias=None):
        """Selects only published pages."""
        queryset = super(PageManager, self).select_published(queryset)
        now = timezone.now()
        # Perform local filtering.
        queryset = queryset.filter(Q(publication_date=None) | Q(publication_date__lte=now))
        queryset = queryset.filter(Q(expiry_date=None) | Q(expiry_date__gt=now))
        # Perform parent ordering.
        quote_name = connection.ops.quote_name
        page_alias = page_alias  or quote_name("pages_page")
        queryset = queryset.extra(
            where = ("""
                NOT EXISTS (
                    SELECT *
                    FROM {pages_page} AS {ancestors}
                    WHERE
                        {ancestors}.{left} < {page_alias}.{left} AND
                        {ancestors}.{right} > {page_alias}.{right} AND (
                            {ancestors}.{is_online} = FALSE OR
                            {ancestors}.{publication_date} > %s OR
                            {ancestors}.{expiry_date} <= %s
                        )
                )
            """.format(
                page_alias = page_alias,
                **dict(
                    (name, quote_name(name))
                    for name in (
                        "pages_page",
                        "ancestors",
                        "left",
                        "right",
                        "is_online",
                        "publication_date",
                        "expiry_date",
                    )
                )
            ),),
            params = (now, now),
        )
        return queryset

    def get_homepage(self):
        """Returns the site homepage."""
        return self.prefetch_related("child_set__child_set").get(parent=None)


class Page(PageBase):

    """A page within the site."""

    objects = PageManager()

    # Hierarchy fields.

    parent = models.ForeignKey(
        "self",
        blank = True,
        null = True,
        default = get_default_page_parent,
        related_name = "child_set",
    )

    left = models.IntegerField(
        editable = False,
        db_index = True,
    )

    right = models.IntegerField(
        editable = False,
        db_index = True,
    )

    @cached_property
    def children(self):
        """The child pages for this page."""
        children = []
        if self.right - self.left > 1:  # Optimization - don't fetch children we know aren't there!
            for child in self.child_set.all():
                child.parent = self
                children.append(child)
        return children

    @property
    def navigation(self):
        """The sub-navigation of this page."""
        return [child for child in self.children if child.in_navigation]

    # Publication fields.

    publication_date = models.DateTimeField(
        blank = True,
        null = True,
        db_index = True,
        help_text = "The date that this page will appear on the website.  Leave this blank to immediately publish this page.",
    )

    expiry_date = models.DateTimeField(
        blank = True,
        null = True,
        db_index = True,
        help_text = "The date that this page will be removed from the website.  Leave this blank to never expire this page.",
    )

    # Navigation fields.

    in_navigation = models.BooleanField(
        "add to navigation",
        default = True,
        help_text = "Uncheck this box to remove this content from the site navigation.",
    )

    # Content fields.

    content_type = models.ForeignKey(
        ContentType,
        editable = False,
        help_text="The type of page content.",
    )

    cached_url = models.CharField(
        max_length=1000,
        null=True,
        blank=True,
    )

    @cached_property
    def content(self):
        """The associated content model for this page."""
        content_cls = ContentType.objects.get_for_id(self.content_type_id).model_class()
        content = content_cls._default_manager.get(page=self)
        content.page = self
        return content

    def reverse(self, view_func, args=None, kwargs=None):
        """Performs a reverse URL lookup."""
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        urlconf = ContentType.objects.get_for_id(self.content_type_id).model_class().urlconf
        return self.get_absolute_url() + urlresolvers.reverse(view_func, args=args, kwargs=kwargs, urlconf=urlconf, prefix="")

    # Standard model methods.

    def get_absolute_url(self, cached=True):
        """Generates the absolute url of the page."""
        if self.cached_url and cached:
            return self.cached_url

        if self.parent:
            url = self.parent.get_absolute_url() + self.url_title + "/"
        else:
            url = urlresolvers.get_script_prefix()

        if url != self.cached_url:
            self.cached_url = url
            self.save()

        return url

    # Tree management.

    @property
    def _branch_width(self):
        return self.right - self.left + 1

    def _excise_branch(self):
        """Excises this whole branch from the tree."""
        branch_width = self._branch_width
        Page.objects.filter(left__gte=self.left).update(
            left = F("left") - branch_width,
        )
        Page.objects.filter(right__gte=self.left).update(
            right = F("right") - branch_width,
        )

    def _insert_branch(self):
        """Inserts this whole branch into the tree."""
        branch_width = self._branch_width
        Page.objects.filter(left__gte=self.left).update(
            left = F("left") + branch_width,
        )
        Page.objects.filter(right__gte=self.left).update(
            right = F("right") + branch_width,
        )

    @transaction.atomic
    def save(self, *args, **kwargs):
        """Saves the page."""
        # Lock entire table.
        existing_pages = dict(
            (page["id"], page)
            for page
            in Page.objects.all().select_for_update().values("id", "parent_id", "left", "right")
        )
        if self.left is None or self.right is None:
            # This page is being inserted.
            if existing_pages:
                parent_right = existing_pages[self.parent_id]["right"]
                # Set the model left and right.
                self.left = parent_right
                self.right = self.left + 1
                # Update the whole tree structure.
                self._insert_branch()
            else:
                # This is the first page to be created, ever!
                self.left = 1
                self.right = 2
        else:
            # This is an update.
            if self.id not in existing_pages:
                old_parent_id = -1
            else:
                old_parent_id = existing_pages[self.id]["parent_id"]

            if old_parent_id != self.parent_id:
                # The page has moved.
                branch_width = self.right - self.left + 1
                # Disconnect child branch.
                if branch_width > 2:
                    Page.objects.filter(left__gt=self.left, right__lt=self.right).update(
                        left = F("left") * -1,
                        right = F("right") * -1,
                    )
                self._excise_branch()
                # Store old left and right values.
                old_left = self.left
                old_right = self.right
                # Put self into the tree.
                parent_right = existing_pages[self.parent_id]["right"]
                if parent_right > self.right:
                    parent_right -= self._branch_width
                self.left = parent_right
                self.right = self.left + branch_width - 1
                self._insert_branch()
                # Put all children back into the tree.
                if branch_width > 2:
                    child_offset = self.left - old_left
                    Page.objects.filter(left__lt=-old_left, right__gt=-old_right).update(
                        left = (F("left") - child_offset) * -1,
                        right = (F("right") - child_offset) * -1,
                    )
        # Now actually save it!
        super(Page, self).save(*args, **kwargs)

        self.get_absolute_url(False)

    def delete(self, *args, **kwargs):
        """Deletes the page."""
        list(Page.objects.all().select_for_update().values_list("left", "right"))  # Lock entire table.
        super(Page, self).delete(*args, **kwargs)
        # Update the entire tree.
        self._excise_branch()

    class Meta:
        unique_together = (("parent", "url_title",),)
        ordering = ("left",)


externals.historylinks("register", Page)


class PageSitemap(sitemaps.PageBaseSitemap):

    """Sitemap for page models."""

    model = Page

    def items(self):
        """Only lists items that are marked as indexable."""
        return filter_indexable_pages(super(PageSitemap, self).items())


sitemaps.register(Page, sitemap_cls=PageSitemap)


class PageSearchAdapter(PageBaseSearchAdapter):

    """Search adapter for Page models."""

    def get_content(self, obj):
        """Returns the search text for the page."""
        content_obj = obj.content

        content = ''

        # Does this object have a section_set?
        section_set = getattr(content_obj.page, 'section_set', None)

        if section_set:
            sections = section_set.all()

            for section in sections:
                field_names = (
                    field.name for field
                    in section._meta.fields
                    if isinstance(field, (models.CharField, models.TextField))
                )

                content = content + self.prepare_content(" ".join(
                    force_text(self._resolve_field(section, field_name))
                    for field_name in field_names
                ))


        return u" ".join((
            super(PageSearchAdapter, self).get_content(obj),
            self.prepare_content(u" ".join(
                unicode(self._resolve_field(content_obj, field_name))
                for field_name in (
                    field.name for field
                    in content_obj._meta.fields
                    if isinstance(field, (models.CharField, models.TextField))
                )
            ))
        )) + content

    def get_live_queryset(self):
        """Selects the live page queryset."""
        # HACK: Prevents a table name collision in the Django queryset manager.
        with publication_manager.select_published(False):
            qs = Page._base_manager.all()
        if publication_manager.select_published_active():
            qs = Page.objects.select_published(qs, page_alias="U0")
        # Filter out unindexable pages.
        qs = filter_indexable_pages(qs)
        # All done!
        return qs


externals.watson("register", Page, adapter_cls=PageSearchAdapter)


# Base content class.

def get_registered_content():
    """Returns a list of all registered content objects."""
    return [
        model for model in models.get_models()
        if issubclass(model, ContentBase) and not model._meta.abstract
    ]


def filter_indexable_pages(queryset):
    """
    Filters the given queryset of pages to only contain ones that should be
    indexed by search engines.
    """
    return queryset.filter(
        robots_index = True,
        content_type__in = [
            ContentType.objects.get_for_model(content_model)
            for content_model
            in get_registered_content()
            if content_model.robots_index
        ]
    )


class ContentBase(models.Model):

    """Base class for page content."""

    # This must be a 64 x 64 pixel image.
    icon = "pages/img/content.png"

    # The heading that the admin places this content under.
    classifier = "content"

    # The urlconf used to power this content's views.
    urlconf = "cms.apps.pages.urls"

    # A fieldset definition. If blank, one will be generated.
    fieldsets = None

    # Whether pages of this type should be included in search indexes. (Can still be disabled on a per-page basis).
    robots_index = True

    page = models.OneToOneField(
        Page,
        primary_key = True,
        editable = False,
        related_name = "+",
    )

    def __unicode__(self):
        """Returns a unicode representation."""
        return unicode(self.page)

    class Meta:
        abstract = True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('news', '0006_auto_20141014_1450'),
    ]

    operations = [
        migrations.AlterField(
            model_name='article',
            name='status',
            field=models.CharField(default=b'draft', max_length=100, choices=[(b'draft', b'Draft'), (b'submitted', b'Submitted for approval'), (b'approved', b'Approved')]),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import cms.apps.news.models
import cms.apps.media.models
from django.conf import settings
import django.db.models.deletion
import cms.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('news', '0001_initial'),
        ('pages', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('media', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='NewsFeed',
            fields=[
                ('page', models.OneToOneField(related_name=b'+', primary_key=True, serialize=False, editable=False, to='pages.Page')),
                ('content_primary', cms.models.fields.HtmlField(verbose_name=b'primary content', blank=True)),
                ('per_page', models.IntegerField(default=5, null=True, verbose_name=b'articles per page', blank=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='category',
            unique_together=set([('url_title',)]),
        ),
        migrations.AddField(
            model_name='article',
            name='authors',
            field=models.ManyToManyField(to=settings.AUTH_USER_MODEL, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='article',
            name='categories',
            field=models.ManyToManyField(to='news.Category', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='article',
            name='image',
            field=cms.apps.media.models.ImageRefField(related_name=b'+', on_delete=django.db.models.deletion.PROTECT, blank=True, to='media.File', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='article',
            name='news_feed',
            field=models.ForeignKey(default=cms.apps.news.models.get_default_news_feed, to='news.NewsFeed'),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='article',
            unique_together=set([('news_feed', 'date', 'url_title')]),
        ),
    ]



"""URL config for {{ project_name }} project."""

from django.conf import settings
from django.conf.urls import patterns, url, include
from django.contrib import admin
from django.views import generic
from django.conf.urls.static import static

from cms.sitemaps import registered_sitemaps
from cms.views import TextTemplateView
from cms.forms import CMSPasswordChangeForm


admin.autodiscover()


urlpatterns = patterns(
    "",

    # Admin URLs.
    url(r'^admin/password_change/$', 'django.contrib.auth.views.password_change',
        {'password_change_form': CMSPasswordChangeForm}, name='password_change'),
    url(r'^admin/password_change/done/$', 'django.contrib.auth.views.password_change_done', name='password_change_done'),
    url(r"^admin/", include(admin.site.urls)),

    # Permalink redirection service.
    url(r"^r/(?P<content_type_id>\d+)-(?P<object_id>[^/]+)/$", "django.contrib.contenttypes.views.shortcut", name="permalink_redirect"),

    # Google sitemap service.
    url(r"^sitemap.xml$", "django.contrib.sitemaps.views.index", {"sitemaps": registered_sitemaps}),
    url(r"^sitemap-(?P<section>.+)\.xml$", "django.contrib.sitemaps.views.sitemap", {"sitemaps": registered_sitemaps}),

    # Basic robots.txt.
    url(r"^robots.txt$", TextTemplateView.as_view(template_name="robots.txt")),

    # There's no favicon here!
    url(r"^favicon.ico$", generic.RedirectView.as_view()),

) + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)


if settings.DEBUG:
    urlpatterns += patterns(
        "",
        url("^404/$", generic.TemplateView.as_view(template_name="404.html")),
        url("^500/$", generic.TemplateView.as_view(template_name="500.html")),
    )


handler500 = "cms.views.handler500"



""""Content used by the links application."""

from django.db import models

from cms.models import LinkField
from cms.apps.pages.models import ContentBase


class Link(ContentBase):

    """A redirect to another URL."""

    classifier = "utilities"

    icon = "links/img/link.png"

    urlconf = "cms.apps.links.urls"

    robots_index = False

    link_url = LinkField(
        "link URL",
        help_text = "The URL where the user will be redirected.",
    )

    new_window = models.BooleanField(
        help_text = "Open the page in a new window.",
        default=False,
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import cms.apps.pages.models


class Migration(migrations.Migration):

    dependencies = [
        ('pages', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='page',
            name='parent',
            field=models.ForeignKey(related_name=b'child_set', default=cms.apps.pages.models.get_default_page_parent, blank=True, to='pages.Page', null=True),
        ),
    ]



"""Debug helpers."""

import traceback
from functools import wraps

from django.conf import settings


def print_exc(func):
    """
    Decorator that prints any raised exceptions to stdout, but only when in debug mode.
    
    This is useful for catching those pesky errors that Django would otherwise squeltch.
    """
    if settings.DEBUG:
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                traceback.print_exc()
                raise
        return wrapper
    else:
        return func
        
        
def print_current_exc():
    """Prints the current exception, if in debug model"""
    if settings.DEBUG:
        traceback.print_exc()


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('news', '0002_auto_20140909_1115'),
    ]

    operations = [
        migrations.AddField(
            model_name='article',
            name='approved',
            field=models.BooleanField(default=False),
            preserve_default=True,
        ),
    ]
